
/*
 * expression.cpp
 *
 * (c) 2006, 2007 by Moritz Beutel
 *
 * Klasse fr den Umgang mit mathematischen Formeln
 * Untersttzt Evaluationen und Ableitungen
 * Entwickelt in OWBasic 5.2, portiert nach C++
 *
 */

#ifndef _UCL_MATH
#define _UCL_MATH

#include <string>
#include <vector>
#include <map>

namespace math
{


class Exception
{
public:
    enum error_t
    {
        SYNTAX = 0,
        NOT_A_CONSTANT,
        INVALID_FUNCTION,
        DOMAIN,
        DIVISION_BY_0,
    };

private:
    error_t err;
    std::string msg;

public:
    Exception (error_t error, const std::string& message)
     : err (error), msg (message) {}
    const std::string& getMessage (void) { return msg; }
    error_t getError (void) { return err; }
};

std::string doubleToString (double d, const char* format = "%g");
std::string intToString (int i, unsigned radix = 10);

class Expression
{
public:
    typedef double (*binary_func_t) (double, double);
    typedef double (*unary_func_t) (double);
    struct func_struct_t
    {
        const char*   name;
        unsigned char priority;
        bool          alpha;    // true - Funktionsname, false - Operator
        const char*   derivation;
    };
    struct binary_struct_t
    {
        binary_func_t func;
        func_struct_t fs;
    };
    struct unary_struct_t
    {
        unary_func_t  func;
        func_struct_t fs;
    };


    static const binary_struct_t binary_functions[];
    static const unary_struct_t unary_functions[];

    static const unsigned FUNCTION_COUNT;
    static const unsigned BINARY_FUNCTION_COUNT;
    static const unsigned UNARY_FUNCTION_COUNT;
    static const unsigned UNARY_FUNCTION_OFFSET;

    static const unsigned TMASK = 0xE0000000,
                          NMASK = ~TMASK;
    static const unsigned TNUM  = 0x20000000,
                          TOPER = 0x40000000,
                          TSUB  = 0x60000000,
                          TSUBD = 0x80000000,
                          TPAR  = 0xA0000000,
                          TVAR  = 0xC0000000,
                          TEND  = 0x00000000;

    static const unsigned EVAL_PASS = 3;
    static const bool     BIND_NEGATIVE = true;


    struct transferitem_t
    {
        unsigned from;
        unsigned to;
        unsigned dst;

        transferitem_t (void) {}
        transferitem_t (unsigned dst_, unsigned from_, unsigned to_)
         : from (from_), to (to_), dst (dst_) {}
    };

    enum callback_task_t
    {
        CT_EVAL,
        CT_EVAL_PAREN,
        CT_EVAL_SIMPLIFY,
        CT_EVAL_SIMPLIFY_STEP, // param
        CT_DERIVE,
        CT_DERIVE_STEP,        // param
        CT_TOTEXT,
        CT_FROMTEXT,
    };
    typedef void (*callback_t) (callback_task_t, unsigned, void*);
private:
    struct par_t
    {
        std::string name;
        int         valoffs;

        par_t (void)
         : valoffs (0) {}
        par_t (const std::string& nm, unsigned offs)
         : name (nm), valoffs (offs) {}
    };


    callback_t cbfunc;
    void* cbval;

    typedef std::vector <unsigned> term_t;
    typedef std::multimap <int, int> map_t;

    std::vector <double> fstk;
    std::vector <transferitem_t> trstk;
    std::vector <par_t> parstk;
    term_t term;
    std::vector <term_t> drbuf;
    map_t mapbuf;

    unsigned fstk_defsize, parstk_defsize;

    std::string varnm;
    int vari;

    unsigned registerDoubleVal (double val);

    static const func_struct_t& getFuncStruct (unsigned idx)
    {
        return ((idx >= UNARY_FUNCTION_OFFSET)
              ? (unary_functions[idx - UNARY_FUNCTION_OFFSET].fs)
              : (binary_functions[idx].fs));
    }

    void stkInit (void);

    void textToTerm (const std::string& text, std::vector <unsigned>& trm);

    void fillBuffer (const term_t& trm);
    void moveBufferIndices (int i, int diff);


    bool isConstantOperand (unsigned op);
    double getConstantValue (unsigned op);

    static int compareOperators (unsigned op1, unsigned op2);
    static void findOperands (const term_t& trm, unsigned i,
                               int& dop, int& lop, int dir);
    static bool isRepOperator (const term_t& trm, unsigned i, unsigned rel_op);
    static int getSurrPriority (const term_t& trm, unsigned i);
    static void removeParen (term_t::iterator p1, term_t::iterator p2,
                             unsigned pr, unsigned pi, unsigned og, bool rv);
    static void evaluateParen (term_t& trm, unsigned i, unsigned pr);
    void doEval (term_t& trm);
    void doDerive (term_t& trm);

    void termDelete (term_t& trm, unsigned i, int n);
    void termExpand (term_t& base, unsigned i, term_t::const_iterator src, unsigned sz);

    void calculateOperator (term_t& trm, unsigned li, unsigned ri, bool brkrepl);
    void evaluateOperator (term_t& base, int i);

    void doCallback (callback_task_t task, unsigned val = 0)
    {
        if (cbfunc)
            cbfunc (task, val, cbval);
    }

public:
    Expression (void);

    unsigned registerParam (const std::string& name);
    void registerParamValue (const std::string& name, double val);
    void removeParamValue (const std::string& name);

    void setVariableName (const std::string& name) { varnm = name; }
    void registerVariableValue (double val) { fstk[0] = val; vari = 0; }
    void removeVariableValue (void) { vari = -1; }

    void registerCallbackFunction (callback_t func, void* val);

    void readFromText (const std::string& text);
    std::string toText (const char* format = "%g");
    double toDouble (void);
    bool isConstantValue (void);

    Expression getEvaluation (void);
    Expression getDerivation (void);

    void evaluate (void);
    void derive (void);

    const std::vector <unsigned>& getTerm (void) { return term; }
};


} // namespace math

#endif /* _UCL_MATH */

